import os
import unittest
import sys

import autograder.fileop
import autograder.util.dir
import autograder.util.file

ALREADY_EXISTS_DIRNAME = "already_exists"
ALREADY_EXISTS_FILENAME = "already_exists.txt"
ALREADY_EXISTS_FILE_POSIX_RELPATH = ALREADY_EXISTS_DIRNAME + "/" + ALREADY_EXISTS_FILENAME
ALREADY_EXISTS_FILE_RELPATH = os.path.join(ALREADY_EXISTS_DIRNAME, ALREADY_EXISTS_FILENAME)

class TestFileOp(unittest.TestCase):
    @unittest.skipIf(sys.platform.startswith("win"), "filops reqire POSIX")
    def test_fileop_validation(self):
        test_cases = [
            # Base
            (["copy", "a", "b"], ["copy", "a", "b"], None),
            (["cp", "a", "b"], ["copy", "a", "b"], None),
            (["move", "a", "b"], ["move", "a", "b"], None),
            (["mv", "a", "b"], ["move", "a", "b"], None),
            (["make-dir", "a"], ["make-dir", "a"], None),
            (["mkdir", "a"], ["make-dir", "a"], None),
            (["remove", "a"], ["remove", "a"], None),
            (["rm", "a"], ["remove", "a"], None),

            # Casing
            (["Copy", "a", "b"], ["copy", "a", "b"], None),
            (["COPY", "a", "b"], ["copy", "a", "b"], None),
            (["CoPy", "a", "b"], ["copy", "a", "b"], None),

            # Normalize Paths
            (["copy", "c/../a", "b"], ["copy", "a", "b"], None),
            (["copy", "a/", "b"], ["copy", "a", "b"], None),
            (["copy", "a//b", "b"], ["copy", "a/b", "b"], None),
            (["copy", "./a", "b"], ["copy", "a", "b"], None),

            # Errors

            # Empty
            (None, None, "File operation is None"),
            ([], None, "File operation is empty"),

            # Number of Args
            (["copy", "a"], None, "Incorrect number of arguments"),
            (["copy", "a", "b", "c"], None, "Incorrect number of arguments"),
            (["move", "a"], None, "Incorrect number of arguments"),
            (["move", "a", "b", "c"], None, "Incorrect number of arguments"),
            (["make-dir"], None, "Incorrect number of arguments"),
            (["make-dir", "a", "b"], None, "Incorrect number of arguments"),
            (["remove"], None, "Incorrect number of arguments"),
            (["remove", "a", "b"], None, "Incorrect number of arguments"),

            # Unknown Command
            (["zzz", "a", "b"], None, "Unknown file operation"),

            # Path Errors
            (["copy", "a\\b", "b"], None, "contains a backslash"),
            (["copy", "/a", "b"], None, "is an absolute path"),
            (["copy", "..", "b"], None, "points outside of the its base directory"),
            (["copy", "../a", "b"], None, "points outside of the its base directory"),
            (["copy", "a/../..", "b"], None, "points outside of the its base directory"),
            (["copy", "a/../../b", "b"], None, "points outside of the its base directory"),
            (["copy", ".", "b"], None, "cannot point just to the current directory"),
            (["copy", "a/..", "b"], None, "cannot point just to the current directory"),
        ]

        for i in range(len(test_cases)):
            with self.subTest(msg = f"Case {i}"):
                (operation, expected, error_substring) = test_cases[i]

                try:
                    autograder.fileop.validate(operation)
                except Exception as ex:
                    if (error_substring is None):
                        self.fail(f"Unexpected error: '{str(ex)}'.")

                    self.assertIn(error_substring, str(ex), 'Error is not as expected.')
                    continue

                if (error_substring is not None):
                    self.fail(f"Did not get expected error: '{error_substring}'.")

                self.assertListEqual(operation, expected, 'Operation not as expected.')

    @unittest.skipIf(sys.platform.startswith("win"), "filops reqire POSIX")
    def test_fileop_exec_copy(self):
        test_cases = [
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, "a", None),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, "a.txt", None),
            (ALREADY_EXISTS_DIRNAME, "a", None),
            (ALREADY_EXISTS_DIRNAME, "a.txt", None),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, ALREADY_EXISTS_DIRNAME + "/a", None),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, "a/b", None),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, ALREADY_EXISTS_DIRNAME, None),
            ("a", "b", "No such file or directory"),
            (ALREADY_EXISTS_DIRNAME, ALREADY_EXISTS_FILE_POSIX_RELPATH, "File exists"),
        ]

        for i in range(len(test_cases)):
            with self.subTest(msg = f"Case {i}"):
                (source, dest, error_substring) = test_cases[i]

                operation = ["cp", source, dest]

                def post_check(operation, temp_dir):
                    expected_source = os.path.normpath(os.path.join(temp_dir, source))
                    expected_dest = os.path.normpath(os.path.join(temp_dir, dest))

                    self.assertTrue(os.path.exists(expected_dest),
                        f"Dest does not exist '{expected_dest}'.")

                    self.assertTrue(os.path.exists(expected_source),
                        f"Source does not exist '{expected_source}'.")

                self._run_fileop_exec_test(operation, error_substring, post_check)

    @unittest.skipIf(sys.platform.startswith("win"), "filops reqire POSIX")
    def test_fileop_exec_move(self):
        test_cases = [
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, "a", None),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, "a.txt", None),
            (ALREADY_EXISTS_DIRNAME, "a", None),
            (ALREADY_EXISTS_DIRNAME, "a.txt", None),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, ALREADY_EXISTS_DIRNAME + "/a", None),
            ("a", "b", "No such file or directory"),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, "a/b", "No such file or directory"),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, ALREADY_EXISTS_DIRNAME, "already exists"),
            (ALREADY_EXISTS_DIRNAME, ALREADY_EXISTS_FILE_POSIX_RELPATH, "into itself"),
        ]

        for i in range(len(test_cases)):
            with self.subTest(msg = f"Case {i}"):
                (source, dest, error_substring) = test_cases[i]

                operation = ["mv", source, dest]

                def post_check(operation, temp_dir):
                    expected_source = os.path.normpath(os.path.join(temp_dir, source))
                    expected_dest = os.path.normpath(os.path.join(temp_dir, dest))

                    self.assertTrue(os.path.exists(expected_dest),
                        f"Dest does not exist '{expected_dest}'.")

                    self.assertFalse(os.path.exists(expected_source),
                        f"Source exists '{expected_source}'.")

                self._run_fileop_exec_test(operation, error_substring, post_check)

    @unittest.skipIf(sys.platform.startswith("win"), "filops reqire POSIX")
    def test_fileop_exec_mkdir(self):
        test_cases = [
            ("a", None),
            ("a/b", None),
            ("a/../b", None),
            (ALREADY_EXISTS_DIRNAME, None),
            (ALREADY_EXISTS_DIRNAME + "/a", None),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, "File exists"),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH + "/a", "Not a directory"),
        ]

        for i in range(len(test_cases)):
            with self.subTest(msg = f"Case {i}"):
                (path, error_substring) = test_cases[i]

                operation = ["mkdir", path]

                def post_check(operation, temp_dir):
                    expected_path = os.path.normpath(os.path.join(temp_dir, path))

                    self.assertTrue(os.path.isdir(expected_path),
                        f"Target does not exist or is not a dir '{expected_path}'.")

                self._run_fileop_exec_test(operation, error_substring, post_check)

    @unittest.skipIf(sys.platform.startswith("win"), "filops reqire POSIX")
    def test_fileop_exec_remove(self):
        test_cases = [
            ("a", None),
            ("a/b", None),
            ("a/../b", None),
            (ALREADY_EXISTS_DIRNAME, None),
            (ALREADY_EXISTS_DIRNAME + "/a", None),
            (ALREADY_EXISTS_FILE_POSIX_RELPATH, None),
        ]

        for i in range(len(test_cases)):
            with self.subTest(msg = f"Case {i}"):
                (path, error_substring) = test_cases[i]

                operation = ["rm", path]

                def post_check(operation, temp_dir):
                    expected_path = os.path.normpath(os.path.join(temp_dir, path))

                    self.assertFalse(os.path.exists(expected_path),
                        f"Target exists '{expected_path}'.")

                self._run_fileop_exec_test(operation, error_substring, post_check)

    def _run_fileop_exec_test(self, operation, error_substring, post_exec):
        autograder.fileop.validate(operation)

        temp_dir = autograder.util.dir.get_temp_dir(prefix = "ag-py-testing-fileop-execute-")

        # Make some existing entries.
        autograder.util.dir.mkdir(os.path.join(temp_dir, ALREADY_EXISTS_DIRNAME))
        autograder.util.file.write(os.path.join(temp_dir, ALREADY_EXISTS_FILE_RELPATH), "AAA")

        try:
            autograder.fileop.execute(operation, temp_dir)
        except Exception as ex:
            if (error_substring is None):
                self.fail(f"Unexpected error: '{str(ex)}'.")

            self.assertIn(error_substring, str(ex), 'Error is not as expected.')
            return

        if (error_substring is not None):
            self.fail(f"Did not get expected error: '{error_substring}'.")

        post_exec(operation, temp_dir)



"""
File Operations (file ops) are simple operations on files
intended to be easy to implement on any platform.

See:
 - https://github.com/edulinq/autograder-server/blob/main/docs/types.md#file-operation-fileop
 - https://github.com/edulinq/autograder-server/blob/main/internal/util/fileop.go
"""

import os

import autograder.util.dir
import autograder.util.dirent
import autograder.util.path

FILE_OP_LONG_COPY = "copy"
FILE_OP_SHORT_COPY = "cp"

FILE_OP_LONG_MOVE = "move"
FILE_OP_SHORT_MOVE = "mv"

FILE_OP_LONG_MKDIR = "make-dir"
FILE_OP_SHORT_MKDIR = "mkdir"

FILE_OP_LONG_REMOVE = "remove"
FILE_OP_SHORT_REMOVE = "rm"

# The long name is the canonical name.
fileop_normal_name = {
    FILE_OP_SHORT_COPY: FILE_OP_LONG_COPY,
    FILE_OP_LONG_COPY: FILE_OP_LONG_COPY,

    FILE_OP_SHORT_MOVE: FILE_OP_LONG_MOVE,
    FILE_OP_LONG_MOVE: FILE_OP_LONG_MOVE,

    FILE_OP_SHORT_MKDIR: FILE_OP_LONG_MKDIR,
    FILE_OP_LONG_MKDIR: FILE_OP_LONG_MKDIR,

    FILE_OP_SHORT_REMOVE: FILE_OP_LONG_REMOVE,
    FILE_OP_LONG_REMOVE: FILE_OP_LONG_REMOVE,
}

# The number of operations for each file operation.
fileop_num_args = {
    FILE_OP_LONG_COPY: 2,
    FILE_OP_LONG_MOVE: 2,
    FILE_OP_LONG_MKDIR: 1,
    FILE_OP_LONG_REMOVE: 1,
}

def validate(operation):
    if (operation is None):
        raise ValueError("File operation is None.")

    if (len(operation) == 0):
        raise ValueError("File operation is empty.")

    command = fileop_normal_name.get(operation[0].lower(), None)
    if (command is None):
        raise ValueError(f"Unknown file operation: '{operation[0]}'.")

    operation[0] = command

    num_args = (len(operation) - 1)
    expected_num_args = fileop_num_args[command]
    if (expected_num_args != num_args):
        raise ValueError(f"Incorrect number of arguments for '{command}' file operation."
            + f" Expected {expected_num_args}, found {num_args}.")

    # Check all path arguments.
    for i in range(1, len(operation)):
        path = operation[i]

        if ('\\' in path):
            raise ValueError(f"Argument at index {i} ('{operation[i]}')"
                + " contains a backslash ('\\') or is not a POSIX path.")

        path = os.path.normpath(path)

        if (os.path.isabs(path)):
            raise ValueError(f"Argument at index {i} ('{operation[i]}') is an absolute path."
                + " Only relative paths are allowed.")

        if (not autograder.util.path.is_local(path)):
            raise ValueError(f"Argument at index {i} ('{operation[i]}')"
                + " points outside of the its base directory."
                + " File operation paths can not reference parent directories.")

        if (path == "."):
            raise ValueError(f"Argument at index {i} ('{operation[i]}')"
                + " cannot point just to the current directory."
                + " File operation paths must point to a"
                + " dirent inside the current directory tree.")

        operation[i] = path

    return operation

# Execute operation operation in the given directory.
def execute(operation, base_dir):
    command = operation[0]

    if (command == FILE_OP_LONG_COPY):
        source_path = _resolve_path(operation[1], base_dir)
        dest_path = _resolve_path(operation[2], base_dir)
        if (source_path == dest_path):
            return

        autograder.util.dirent.copy(source_path, dest_path)
    elif (command == FILE_OP_LONG_MOVE):
        source_path = _resolve_path(operation[1], base_dir)
        dest_path = _resolve_path(operation[2], base_dir)
        if (source_path == dest_path):
            return

        autograder.util.dirent.move(source_path, dest_path)
    elif (command == FILE_OP_LONG_MKDIR):
        path = _resolve_path(operation[1], base_dir)

        autograder.util.dir.mkdir(path)
    elif (command == FILE_OP_LONG_REMOVE):
        path = _resolve_path(operation[1], base_dir)
        autograder.util.dirent.remove(path)
    else:
        raise ValueError(f"Unknown file operation: '{command}'.")

def validate_file_operations(operations):
    for operation in operations:
        validate(operation)

def exec_file_operations(operations, base_dir):
    for operation in operations:
        execute(operation, base_dir)

def _resolve_path(path, base_dir):
    if (os.path.isabs(path)):
        return os.path.normpath(path)

    return os.path.normpath(os.path.join(base_dir, path))



import os

import autograder.util.dirent

def get_temp_dir(**kwargs):
    path = autograder.util.dirent.get_temp_path(**kwargs)
    mkdir(path)
    return path

def mkdir(path):
    os.makedirs(path, exist_ok = True)



import os

THIS_DIR = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))

# Reserved filenames (mainly on Windows).
# See: https://github.com/python/cpython/blob/3.13/Lib/ntpath.py
RESERVED_NAMES = frozenset(
    {'CON', 'PRN', 'AUX', 'NUL', 'CONIN$', 'CONOUT$'} |  # noqa: W504
    {f'COM{c}' for c in '123456789\xb9\xb2\xb3'} |  # noqa: W504
    {f'LPT{c}' for c in '123456789\xb9\xb2\xb3'}
)

# Try to see of a file is "local" using only lexical analysis.
# See: https://pkg.go.dev/path/filepath#IsLocal
# This is not robust and should not be used for any systems where security can be an issue.
def is_local(path):
    path = os.path.normpath(path.strip())

    if (path == ''):
        return False

    if (os.path.isabs(path)):
        return False

    if (_is_reserved(path)):
        return False

    # To see if the path does not break out of the current directory,
    # join it to the current directory and see if the current
    # directory is still a base path.

    current_path = THIS_DIR

    # Note that the normalization should remove any (or several)
    # parent references ('..') at this point.
    test_path = os.path.normpath(os.path.join(THIS_DIR, path))

    common_prefix = os.path.commonprefix([current_path, test_path])
    if (common_prefix != current_path):
        return False

    return True

# A very weak version of ntpath.isreserved(path) (which was added in version 3.13).
# https://docs.python.org/3/library/os.path.html#os.path.isreserved
def _is_reserved(path):
    return os.path.basename(path).upper() in RESERVED_NAMES



import json
import sys

import autograder.api.courses.assignments.submissions.analysis.individual

def run(arguments):
    result = autograder.api.courses.assignments.submissions.analysis.individual.send(
        arguments, exit_on_error = True)

    print(json.dumps(result, indent = 4))
    return 0

def main():
    return run(_get_parser().parse_args())

def _get_parser():
    parser = autograder.api.courses.assignments.submissions.analysis.individual._get_parser()
    return parser

if (__name__ == '__main__'):
    sys.exit(main())
